@@ -67,7 +67,8 @@
 #define OPT_EV_EXIT		(0x00000400)
 #define OPT_EV_CORE		(0x00000800)
 #define OPT_EV_COMM		(0x00001000)
-#define OPT_EV_MASK		(0x00001f00)
+#define OPT_EV_CLNE		(0x00002000)
+#define OPT_EV_MASK		(0x00003f00)
 #define OPT_EV_ALL		(OPT_EV_MASK)
 
 #define	GOT_TGID		(0x01)
@@ -81,11 +82,11 @@
 
 /* /proc info cache */
 typedef struct proc_info {
+	struct proc_info *next;	/* next proc info in hashed linked list */
 	pid_t	pid;		/* Process ID */
 	char	*cmdline;	/* /proc/pid/cmdline text */
 	bool	kernel_thread;	/* true if a kernel thread */
 	struct timeval start;	/* time when process started */
-	struct proc_info *next;	/* next proc info in hashed linked list */
 } proc_info_t;
 
 /* For kernel task checking */
@@ -100,14 +101,15 @@ typedef enum {
 	STAT_EXIT,
 	STAT_CORE,
 	STAT_COMM,
+	STAT_CLNE,
 	STAT_LAST
 } event_t;
 
 typedef struct proc_stats {
+	struct proc_stats *next;/* Next one in list */
 	char *name;		/* Process name */
 	uint64_t count[STAT_LAST]; /* Tally count */
 	uint64_t total;		/* Tally count total of all counts */
-	struct proc_stats *next;/* Next one in list */
 } proc_stats_t;
 
 typedef struct {
@@ -124,6 +126,7 @@ static const ev_map_t ev_map[] = {
 	{ "exit", "Exit", OPT_EV_EXIT, STAT_EXIT },
 	{ "core", "Coredump", OPT_EV_CORE, STAT_CORE },
 	{ "comm", "Comm", OPT_EV_COMM, STAT_COMM },
+	{ "clone","Clone", OPT_EV_CLNE, STAT_CLNE },
 	{ "all",  ""	, OPT_EV_ALL,   0 },
 	{ NULL  ,  NULL,   0,           0 }
 };
@@ -416,30 +419,26 @@ static inline int proc_info_hash(const pid_t pid)
 	return pid % MAX_PIDS;
 }
 
+
 /*
  *  proc_name_hash()
- *	hash on proc name, from Aho, Sethi, Ullman, Compiling Techniques.
+ *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
  */
-static inline int proc_name_hash(const char *str)
+static inline unsigned int proc_name_hash(const char *str)
 {
-	unsigned long h = 0;
+	register unsigned int hash = 5381;
+	register unsigned int c;
 
-	while (*str) {
-		unsigned long g;
-		h = (h << 4) + (*str);
-		if (0 != (g = h & 0xf0000000)) {
-			h = h ^ (g >> 24);
-			h = h ^ g;
-		}
-		str++;
+	while ((c = *str++)) {
+		/* (hash * 33) ^ c */
+		hash = ((hash << 5) + hash) ^ c;
 	}
-
-	return h % MAX_PIDS;
+	return hash % MAX_PIDS;
 }
 
 static void proc_stats_account(const pid_t pid, const event_t event)
 {
-	int h;
+	unsigned int h;
 	char *name;
 	proc_stats_t *stats;
 	proc_info_t *info;
@@ -1018,14 +1017,17 @@ static int monitor(const int sock)
 			switch (proc_ev->what) {
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
 			case PROC_EVENT_FORK:
-				proc_stats_account(proc_ev->event_data.fork.parent_pid, STAT_FORK);
+				ppid = get_parent_pid(proc_ev->event_data.fork.child_pid, &is_thread);
+				proc_stats_account(proc_ev->event_data.fork.parent_pid,
+					is_thread ? STAT_CLNE : STAT_FORK);
 				if (gettimeofday(&tv, NULL) < 0) {
 					memset(&tv, 0, sizeof tv);
 				}
-				ppid = get_parent_pid(proc_ev->event_data.fork.child_pid, &is_thread);
 				info1 = proc_info_get(ppid);
 				info2 = proc_info_add(proc_ev->event_data.fork.child_pid, &tv);
-				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_FORK)) {
+				if (!(opt_flags & OPT_QUIET) &&
+					(((opt_flags & OPT_EV_FORK) && !is_thread) ||
+					 ((opt_flags & OPT_EV_CLNE) && is_thread))) {
 					if (info1 != NULL && info2 != NULL) {
 						char *type = is_thread ? "clone" : "fork";
 						row_increment();
@@ -1156,11 +1158,15 @@ static void show_help(char *const argv[])
 	printf("-q\trun quietly and enable -S option.\n");
 }
 
+/*
+ *  parse_ev()
+ *	parse event strings, turn into flag mask
+ */
 static int parse_ev(const char *arg)
 {
-	char *str, *token, *saveptr = NULL;
+	char *str, *token;
 
-	for (str = (char*)arg; (token = strtok_r(str, ",", &saveptr)) != NULL; str = NULL) {
+	for (str = (char*)arg; (token = strtok(str, ",")) != NULL; str = NULL) {
 		int i;
 		bool found = false;
 
@@ -1222,7 +1228,7 @@ int main(int argc, char * const argv[])
 	}
 
 	if ((opt_flags & OPT_EV_MASK) == 0)
-		opt_flags |= (OPT_EV_FORK | OPT_EV_EXEC | OPT_EV_EXIT);
+		opt_flags |= (OPT_EV_FORK | OPT_EV_EXEC | OPT_EV_EXIT | OPT_EV_CLNE);
 
 	if (geteuid() != 0) {
 		fprintf(stderr, "Need to run with root access.\n");
